@@ -135,6 +135,7 @@ enum cmd_and_opt_values {
 
   oPassphraseFD,
   oPinentryMode,
+  oNoProtection,
   oRequestOrigin,
 
   oAssumeArmor,
@@ -217,7 +218,13 @@ enum cmd_and_opt_values {
   oCompatibilityFlags,
   oKbxBufferSize,
   oAlwaysTrust,
-  oNoAutostart
+  oNoAutostart,
+  oAssertSigner,
+  oAssertValidsig,
+  oAssertKeyusage,
+  oNoQESNote,
+
+  oNoop
  };
 
 
@@ -319,7 +326,7 @@ static gpgrt_opt_t opts[] = {
   ARGPARSE_s_s (oKeyboxdProgram, "keyboxd-program", "@"),
   ARGPARSE_s_s (oDirmngrProgram, "dirmngr-program", "@"),
   ARGPARSE_s_s (oProtectToolProgram, "protect-tool-program", "@"),
-
+  ARGPARSE_s_n (oNoQESNote, "no-qes-note", "@"),
 
   ARGPARSE_header ("Input", N_("Options controlling the input")),
 
@@ -432,6 +439,10 @@ static gpgrt_opt_t opts[] = {
   ARGPARSE_s_n (oDisableFdTranslation, "disable-fd-translation", "@"),
   ARGPARSE_s_i (oPassphraseFD,    "passphrase-fd", "@"),
   ARGPARSE_s_s (oPinentryMode,    "pinentry-mode", "@"),
+  ARGPARSE_s_n (oNoProtection,    "no-protection", "@"),
+  ARGPARSE_s_s (oAssertSigner,   "assert-signer", "@"),
+  ARGPARSE_s_n (oAssertValidsig, "assert-validsig", "@"),
+  ARGPARSE_s_s (oAssertKeyusage, "assert-keyusage", "@"),
 
 
   ARGPARSE_header (NULL, N_("Other options")),
@@ -496,12 +507,20 @@ static struct debug_flags_s debug_flags [] =
 static struct compatibility_flags_s compatibility_flags [] =
   {
     { COMPAT_ALLOW_KA_TO_ENCR, "allow-ka-to-encr" },
+    { COMPAT_NO_CHAIN_CACHE, "no-chain-cache"     },
+    { COMPAT_NO_KEYINFO_CACHE, "no-keyinfo-cache" },
     { 0, NULL }
   };
 
 
 /* Global variable to keep an error count. */
 int gpgsm_errors_seen = 0;
+/* If opt.assert_signer_list is used and this variable is not true
+ * gpgsm will be forced to return EXIT_FAILURE.  */
+int assert_signer_true = 0;
+/* If opt.assert_validsig is used and this variable is not true
+ * gpgsm will be forced to return EXIT_FAILURE.  */
+int assert_validsig_true = 0;
 
 /* It is possible that we are currentlu running under setuid permissions */
 static int maybe_setuid = 1;
@@ -529,6 +548,9 @@ static int default_include_certs = DEFAULT_INCLUDE_CERTS;
 /* Whether the chain mode shall be used for validation.  */
 static int default_validation_model;
 
+/* Counter used to convey data from deinit_ctrl to gpgsm_exit.  */
+static unsigned int parent_cache_stats;
+
 /* The default cipher algo.  */
 #define DEFAULT_CIPHER_ALGO "AES256"
 
@@ -763,11 +785,11 @@ set_debug (void)
   else if (!strcmp (debug_level, "none") || (numok && numlvl < 1))
     opt.debug = 0;
   else if (!strcmp (debug_level, "basic") || (numok && numlvl <= 2))
-    opt.debug = DBG_IPC_VALUE;
+    opt.debug = DBG_MEMSTAT_VALUE;
   else if (!strcmp (debug_level, "advanced") || (numok && numlvl <= 5))
-    opt.debug = DBG_IPC_VALUE|DBG_X509_VALUE;
+    opt.debug = DBG_MEMSTAT_VALUE|DBG_X509_VALUE;
   else if (!strcmp (debug_level, "expert")  || (numok && numlvl <= 8))
-    opt.debug = (DBG_IPC_VALUE|DBG_X509_VALUE
+    opt.debug = (DBG_MEMSTAT_VALUE|DBG_X509_VALUE
                  |DBG_CACHE_VALUE|DBG_CRYPTO_VALUE);
   else if (!strcmp (debug_level, "guru") || numok)
     {
@@ -1166,6 +1188,10 @@ main ( int argc, char **argv)
             log_error (_("invalid pinentry mode '%s'\n"), pargs.r.ret_str);
 	  break;
 
+        case oNoProtection:
+          ctrl.no_protection = 1;
+          break;
+
         case oRequestOrigin:
           opt.request_origin = parse_request_origin (pargs.r.ret_str);
           if (opt.request_origin == -1)
@@ -1518,6 +1544,22 @@ main ( int argc, char **argv)
           keybox_set_buffersize (pargs.r.ret_ulong, 0);
           break;
 
+        case oAssertSigner:
+          add_to_strlist (&opt.assert_signer_list, pargs.r.ret_str);
+          break;
+
+        case oAssertValidsig:
+          opt.assert_validsig = 1;
+          break;
+
+        case oAssertKeyusage:
+          add_to_strlist (&opt.assert_keyusage, pargs.r.ret_str);
+          break;
+
+        case oNoQESNote: opt.no_qes_note = 1; break;
+
+        case oNoop: break;
+
         default:
           if (configname)
             pargs.err = ARGPARSE_PRINT_WARNING;
@@ -1616,7 +1658,7 @@ main ( int argc, char **argv)
   assuan_control (ASSUAN_CONTROL_REINIT_SYSCALL_CLAMP, NULL);
 
 
-/*   if (opt.qualsig_approval && !opt.quiet) */
+/*   if (opt.qualsig_approval && !opt.quiet && !opt.no_qes_note) */
 /*     log_info (_("This software has officially been approved to " */
 /*                 "create and verify\n" */
 /*                 "qualified signatures according to German law.\n")); */
@@ -1785,6 +1827,10 @@ main ( int argc, char **argv)
       gnupg_inhibit_set_foregound_window (1);
     }
 
+  /* Better make sure that we have a statusfp so that a failure status
+   * in gpgsm_exit can work even w/o any preeding status messages. */
+  gpgsm_init_statusfp (&ctrl);
+
   /* Add default keybox. */
   if (!nrings && default_keyring && !opt.use_keyboxd)
     {
@@ -1989,15 +2035,10 @@ main ( int argc, char **argv)
         else
           wrong_args ("--sign [datafile]");
 
-#if GPGRT_VERSION_NUMBER >= 0x012700 /* >= 1.39 */
         if (err)
           gpgrt_fcancel (fp);
         else
           es_fclose (fp);
-#else
-        (void)err;
-        es_fclose (fp);
-#endif
       }
       break;
 
@@ -2241,11 +2282,11 @@ main ( int argc, char **argv)
 
 
     case aLearnCard:
-      if (argc)
+      if (argc > 1)
         wrong_args ("--learn-card");
       else
         {
-          int rc = gpgsm_agent_learn (&ctrl);
+          int rc = gpgsm_agent_learn (&ctrl, argc? *argv : NULL);
           if (rc)
             log_error ("error learning card: %s\n", gpg_strerror (rc));
         }
@@ -2329,16 +2370,33 @@ emergency_cleanup (void)
 void
 gpgsm_exit (int rc)
 {
+  if (rc)
+    ;
+  else if (log_get_errorcount(0))
+    rc = 2;
+  else if (gpgsm_errors_seen)
+    rc = 1;
+  else if (opt.assert_signer_list && !assert_signer_true)
+    rc = 1;
+  else if (opt.assert_validsig && !assert_validsig_true)
+    rc = 1;
+
+  /* If we had an error but not printed an error message, do it now.
+   * Note that the function will never print a second failure status
+   * line. */
+  if (rc)
+    gpgsm_exit_failure_status ();
+
   gcry_control (GCRYCTL_UPDATE_RANDOM_SEED_FILE);
   if (opt.debug & DBG_MEMSTAT_VALUE)
     {
+      log_info ("cert_chain_cache: cached=%u\n", parent_cache_stats);
       gcry_control( GCRYCTL_DUMP_MEMORY_STATS );
       gcry_control( GCRYCTL_DUMP_RANDOM_STATS );
     }
   if (opt.debug)
     gcry_control (GCRYCTL_DUMP_SECMEM_STATS );
   emergency_cleanup ();
-  rc = rc? rc : log_get_errorcount(0)? 2 : gpgsm_errors_seen? 1 : 0;
   exit (rc);
 }
 
@@ -2360,9 +2418,23 @@ gpgsm_init_default_ctrl (struct server_control_s *ctrl)
 void
 gpgsm_deinit_default_ctrl (ctrl_t ctrl)
 {
+  unsigned int n;
+
   gpgsm_keydb_deinit_session_data (ctrl);
+  gpgsm_flush_keyinfo_cache (ctrl);
   xfree (ctrl->revocation_reason);
   ctrl->revocation_reason = NULL;
+  n = 0;
+  while (ctrl->parent_cert_cache)
+    {
+      cert_cache_item_t next = ctrl->parent_cert_cache->next;
+      ksba_cert_release (ctrl->parent_cert_cache->result);
+      xfree (ctrl->parent_cert_cache);
+      ctrl->parent_cert_cache = next;
+      n++;
+    }
+  if (n > parent_cache_stats)
+    parent_cache_stats = n;
 }
 
 
